/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/dlist.h
* Description:  Doubly-linked list.
* Author:       Brett Kessler
* Created:      Sun Apr 19 10:00:42 1992
* Modified:     Sun May  3 18:11:41 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#if !defined(DLIST_H)
#define DLIST_H

#include "header.h"
#include "Malloc.h"
#include "resource.h"

/*
 * DList_Cons
 *
 * A dlist is built up from a doubly linked list of Dlist_Cons.  A Dlist_Cons
 * contains a client datum and a pointer to the previos and next Cons'es.  
 * next is NULL at end of list, prev is NULL at beginning.
 */
typedef struct DList_cons {
  struct DList_cons* prev;
  void* datum;
  struct DList_cons* next;
}* DList_Cons;

/*
 * DList
 * 
 * A handle on a dlist.  count tells how many data.  head is first element
 * of list, tail is last element (facilitating adding elements to end).
 * Optional resource is used for making new Cons; if NULL, they will be
 * malloc'ed.
 */
typedef struct dlist {
  unsigned int count;
  DList_Cons head;
  DList_Cons tail;
  Resource consResource;
}* DList;

/*
 * DList_Init
 * 
 * If client allocates its own struct list, e.g., as an automatic variable,
 * it must call this function to initialize the structure.
 */
extern void
DList_Init(DList dlist);

/*
 * DList_Create
 *
 * Creates a new dlist.  Client does not need to call DList_Init.
 */
extern DList
DList_Create(void);

/*
 * DList_Add
 *
 * Adds a new element to the end of the dlist.  If dlist is NULL, creates a 
 * new one.  Returns the dlist.  If the dlist has a consResource, that is
 * used for the new cons; otherwise it will be malloc'ed.
 */
extern DList
DList_Add(DList dlist, void* datum);

/*
 * DList_AddCons
 *
 * Like DList_Add, but returns the new cons.
 */
extern DList_Cons
DList_AddCons(DList dlist, void* datum);

#define DList_DEFAULT_CONS_PER_BLOCK 1000

/*
 * DList_CreateCons
 *
 * Creates a Resource that can be used for allocating DList_Cons elements.
 */
extern Resource
DList_CreateCons(unsigned int consPerBlock);

/*
 * DList_First
 *
 * Returns the first datum on the list, without removing it.
 */
extern void*
DList_First(DList dlist);

/*
 * DList_Comparer
 */
typedef Boolean (*DList_Comparer)(void* datum1, void* datum2);

/*
 * DList_DirectCompare
 *
 * Returns T iff the two pointers are the same address.
 */
extern Boolean
DList_DirectCompare(void* datum1, void* datum2);

/*
 * DList_Equal
 */
extern Boolean
DList_Equal(DList list1, DList list2, DList_Comparer function);

/*
 * DList_Append
 *
 * Appends list2 to the end of list1.
 */
extern void
DList_Append(DList list1, DList list2);

#endif  /* end of code */

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
