/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/factor.c
* Description:  Implements factor.h
* Author:       Brett Kessler
* Created:      Wed Apr 15 20:16:33 1992
* Modified:     Sat May 23 09:26:11 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "factor.h"
#include "alternatives.h"
#include "attributes.h"
#include "segments.h"
#include "item.h"
#include "Malloc.h"
#include <stdarg.h>

Factor
Factor_CreateFromBoundary(Factor_Boundary bound)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_boundaryType;
  f->u.boundary = bound;
  return f;
}

Factor
Factor_CreateWordBoundary(Attributes attributes)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_wordBoundaryType;
  f->u.attributes = attributes;
  return f;
}

Factor
Factor_CreateFromAVM(AVM avm)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_avmType;
  f->u.avmFeatures = avm;
  return f;
}

Factor
Factor_CreateFromAlternative(int nItems, ...)
{
  Factor f;
  Alternatives alts;
  int i;
  va_list args;
  // typedef void* Item;
  va_start(args, nItems);
  f = Malloc_Struct(factor);
  f->type = Factor_alternativeType;
  alts = Malloc_Struct(alternatives);
  f->u.alternatives = alts;
  alts->nItems = nItems;
  alts->items = Malloc(nItems * sizeof (Item));
  for (i = 0; i < nItems; i++)
    alts->items[i] = va_arg(args, Item);
  va_end(args);
  return f;
}


Factor
Factor_CreateFromIndexedAVM(IndexedAVM avm)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_indexedAVMType;
  f->u.indexedAVM = avm;
  return f;
}

Factor
Factor_CreateFromZero(void)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_zeroType;
  return f;
}

Factor
Factor_CreateFromSegment(int id)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_segmentType;
  f->u.number = Segments_SegmentFromID(id);
  return f;
}

Factor
Factor_CreateFromOption(void* item)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_optionalType;
  f->u.optionalItem = item;
  return f;
}

Factor
Factor_CreateFromRepetition(Factor rep)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_repetitionType;
  f->u.repetition = rep;
  return f;
}

Factor
Factor_CreateFocusStart(void)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_startFocusType;
  return f;
}

Factor
Factor_CreateFocusEnd(void)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_endFocusType;
  return f;
}

/*
 * When a focus-start token is encountered in a pattern match, we simply
 * record the current position in the data string.  This plus the focus-end
 * notation below will tell the caller which part of the structural
 * description was the source item which is to be replaced by the structural
 * change.
 */
static void
StartFocus(Match_Facts facts)
{
  facts->out.focusStart = facts->in.text;
  Match_Continue(facts);
}

static void
EndFocus(Match_Facts facts)
{
  facts->out.focusEnd = facts->in.text->prev;
  Match_Continue(facts);
}

/* If the current position in the string is a boundary of the same type,
   push on to the continuation agenda; else return in failure.
 */
static void
MatchBoundary(Factor_Boundary boundary, Match_Facts facts)
{
  Terminal term;
  DList_Cons currentCons;
  Boolean matches;
  currentCons = facts->in.text;
  term = currentCons->datum;
  switch (boundary) {
  case Factor_syllableBoundary:
    matches = (term->type == Terminal_syllableBoundary); break;
  case Factor_utteranceBoundary:
    matches = (term->type == Terminal_utteranceBoundary); break;
  default: 
    (void) fprintf(stderr, "Boundary factor is of no known type.\n");
    exit(1);
  }
  if (matches) {
    facts->in.text = currentCons->next;
    Match_Continue(facts);
    if (facts->out.success) return;
  }
  facts->in.text = currentCons;
}

void
Factor_Push(Factor factor, Match_Facts facts)
{
  Cons newCons;
  newCons = Resource_Get(facts->consResource);
  newCons->datum = factor;
  newCons->next = facts->in.continuation;
  facts->in.continuation = newCons;
}

/*
 * Tries to match the smallest number of iterations of the Factor
 * pattern that will lead to success.  First attempts simply to match
 * the continuation agenda.  If that fails, tries matching the pattern;
 * if that succeeds, advances the string to the end of that match and
 * tries the continuation again.  If that fails, tries matching the
 * pattern twice, etc.  This continues until the continuation succeeds,
 * or the repetition pattern fails.
 */
static void
MatchRepetition(Factor factor, Match_Facts facts)
{
  struct Match_facts subfacts;
  int count = 0;
  Match_Continue(facts);
  if (facts->out.success) return;
  subfacts = *facts;
  subfacts.in.continuation = NULL;
  while (1) {
    Factor_Push(factor, &subfacts);
    subfacts.out.success = FALSE;
    Match_Continue(&subfacts);
    if (subfacts.out.success) {
      facts->in.text = subfacts.out.end->next;
      Match_Continue(facts);
      if (facts->out.success) {
	facts->in.text = subfacts.in.text;
	return;
      }
    } else {
      facts->in.text = subfacts.in.text;
      return;
    }
    if (count > 10) {
      /* Protection against calamitous programming errors. */
      printf("Too many repetitions.\n");
      exit(1);
    }
  }
}

/* Matching a segment (such as bh) succeeds if the current string token
 * is that segment and the continuation agenda succeeds.
 */
static void
MatchSegment(Segment segment, Match_Facts facts)
{
  DList_Cons currentPos;
  Terminal currentTerm;
  currentPos = facts->in.text;
  currentTerm = currentPos->datum;
  if (currentTerm->type != Terminal_segmentType) return;
  if (currentTerm->u.segment != segment) return;
  facts->in.text = currentPos->next;
  Match_Continue(facts);
  facts->in.text = currentPos;
}

/* Matching a word boundary (WB) succeeds if the current string token
 * is a word boundary (" ") with the indicated lexical properties,
 * and the continuation agenda succeeds.
 */
static void
MatchWordBoundary(Attributes attributes, Match_Facts facts)
{
  DList_Cons currentPos;
  Terminal currentTerm;
  int attrsMatch;
  currentPos = facts->in.text;
  currentTerm = currentPos->datum;
  if (currentTerm->type != Terminal_wordBoundary) return;  
  attrsMatch = Attributes_Match(attributes, currentTerm);
  if (!attrsMatch) return; 
  facts->in.text = currentPos->next;
  Match_Continue(facts);
  if (!facts->out.success) facts->in.text = currentPos;
}

void
Factor_Match(Factor factor, Match_Facts facts)
{
  switch (factor->type) {
  case Factor_itemType: 
    Item_Match(factor->u.item, facts); break;
  case Factor_alternativeType: 
    Alternatives_Match(factor->u.alternatives, facts); break;
  case Factor_optionalType: 
    Item_MatchOptional(factor->u.optionalItem, facts); break;
  case Factor_repetitionType: 
    MatchRepetition(factor->u.repetition, facts); break;
  case Factor_avmType: 
    AVM_Match(factor->u.avmFeatures, facts); break;
  case Factor_indexedAVMType: 
    IndexedAVM_Match(factor->u.indexedAVM, facts); break;
  case Factor_segmentType: 
    MatchSegment(factor->u.number, facts); break;
  case Factor_startFocusType: 
    StartFocus(facts); break;
  case Factor_endFocusType: 
    EndFocus(facts); break;
  case Factor_zeroType: 
    Match_Continue(facts); break;
  case Factor_boundaryType: 
    MatchBoundary(factor->u.boundary, facts); break;
  case Factor_wordBoundaryType:
    //fprintf(stderr, "Factor_Match Factor_wordBoundaryType\n");
    MatchWordBoundary(factor->u.attributes, facts); break;
  default:
    (void)
      fprintf(stderr, "Matching unknown kind of factor: %d\n", factor->type);
    exit(1);
  }
}

Factor
Factor_CreateFromItem(void* /* Item */ item)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_itemType;
  f->u.item = item;
  return f;
}

DList_Cons
Factor_InsertAfter(Factor factor,
		   Terminal_String string,
		   DList_Cons afterCons,
		   Match_Facts facts,
		   Segment deletedSegment)
{
  // fprintf(stderr, "Factor_InsertAfter(%X), afterCons=%X, type %d\n", 
// 	  factor, afterCons, factor->type);
  switch (factor->type) {
  case Factor_zeroType: return NULL;
  case Factor_avmType:
    return AVM_InsertAfter(factor->u.avmFeatures,
		    string, 
		    afterCons,
		    facts->in.vars,
		    deletedSegment);
  case Factor_indexedAVMType:
    return IndexedAVM_InsertAfter(factor->u.indexedAVM,
			   string,
			   afterCons,
			   facts->in.indexes,
			   facts->in.vars);
  case Factor_segmentType:
    return Terminal_InsertSegment(string, factor->u.number, afterCons);
  case Factor_boundaryType:
    // fprintf(stderr, "Factor_boundaryType\n");
    return Terminal_InsertBoundary(string, 
			    (factor->u.boundary == Factor_syllableBoundary) ?
			      Terminal_syllableBoundary :
			      Terminal_utteranceBoundary,
			    afterCons);
  default:
    (void) 
      fprintf(stderr, "Don't know how to insert a factor of type %d\n",
	      factor->type);
    exit(1);
  }
  // fprintf(stderr, "Factor_InsertAfter returns\n");
  //Error: we're supposed to return a DList_Cons here!
  //I assume it corresponds to the new item we've just entered.
}

/*
******************************************************************************
*                                End of file
******************************************************************************
*/

