/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/factor.h
* Description:  A top-level constituent of a rule "item" pattern.
* Author:       Brett Kessler
* Created:      Sat Apr 25 09:15:46 1992
* Modified:     Mon May  4 08:24:58 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#if !defined(FACTOR_H)
#define FACTOR_H

#include "alternatives.h" 
#include "attributes.h"
#include "avm.h" 
#include "indexedAvm.h"
#include "match.h"
#include "segment.h"

typedef enum factor_Boundary {
  Factor_syllableBoundary,  /* SB matches "-" */
  Factor_wordBoundary,      /* WB matches " " */
  Factor_utteranceBoundary  /* UB matches "#" */
} Factor_Boundary;

typedef enum factor_Type {
  Factor_itemType,          /* A B C */
  Factor_alternativeType,   /* {A, B, C} */
  Factor_optionalType,      /* (A B) */
  Factor_repetitionType,    /* A* */
  Factor_avmType,           /* [+x -y {-a, +b}] */
  Factor_indexedAVMType,    /* [+x -y {-a, +b}]_i */
  Factor_segmentType,       /* bh */
  Factor_startFocusType,    /* introduces source description */
  Factor_endFocusType,      /* delimits end of source description */
  Factor_zeroType,          /* 0 */
  Factor_boundaryType,      /* SB or UB */
  Factor_wordBoundaryType   /* WB<+x, -y> */
} Factor_Type;
/*
 * Factor
 *
 * A top level component of an Item.  Actually an abstract type which
 * can be any of the things listed here.
 */
typedef struct factor {
  Factor_Type type;
  union { 
    Alternatives alternatives; /* _alternativeType */
    void* /* Item */ optionalItem; /*_optionalType */
    void* /* Item */ item;
    AVM avmFeatures; /* _avmType*/
    struct factor* repetition; /* _repetitionType */ 
    IndexedAVM indexedAVM; 
    Segment number; /* _segmentType */
    Factor_Boundary boundary; /* _boundaryType */
    Attributes attributes; /* _wordBoundaryType */
  } u;
}* Factor;

/*
 * Factor_CreateFromItem
 *
 * Creates a factor which consists of another item.
 */
extern Factor 
Factor_CreateFromItem(void* /* Item */ item);

/*
 * Factor_CreateFromBoundary
 *
 * Creates a Factor for matching a syllable or utterance boundary.
 */
extern Factor 
Factor_CreateFromBoundary(Factor_Boundary bound);

/*
 * Factor_CreateWordBoundary
 *
 * Creates a word boundary for representing the end of a word that has the
 * given lexical attributes.
 */ 
extern Factor
Factor_CreateWordBoundary(Attributes attributes);

/*
 * Factor_CreateFromAVM
 *
 * Creates an AVM factor from an AVM.
 */
extern Factor 
Factor_CreateFromAVM(AVM avm);

/*
 * Factor_CreateFromAlternative
 *
 * Builds an alternative Factor.  nItems tell how many alternatives there
 * are; the variable arguments are that many Item's.
 */
extern Factor 
Factor_CreateFromAlternative(int nItems, ...);

/*
 * Factor_CreateFromIndexedAVM
 *
 * Creates a Factor from an IndexedAVM.
 */
extern Factor 
Factor_CreateFromIndexedAVM(IndexedAVM avm);

/*
 * Factor_CreateFromZero
 *
 * Creates a Factor that matches 0 (always succeeds).
 */
extern Factor 
Factor_CreateFromZero(void);

/*
 * Factor_CreateFromSegment
 *
 * Creates a Factor for matching an exact segment.
 */
extern Factor 
Factor_CreateFromSegment(int segment);

/*
 * Factor_CreateFromOption
 *
 * Creates a Factor that optionally matches the given item.
 */
extern Factor 
Factor_CreateFromOption(void* item);

/*
 * Factor_CreateFromRepetition
 *
 * Creates a Factor that matches any number of the given Factor.
 */
extern Factor 
Factor_CreateFromRepetition(Factor rep);

/*
 * Factor_CreateFocusStart
 *
 * Creates a factor that delimits the start of the source description that
 * has been in-lined for an environment bar.
 */
extern Factor 
Factor_CreateFocusStart(void);

/*
 * Factor_CreateFocusEnd
 *
 * Creates a factor that delimits the end of a source description.
 */
extern Factor 
Factor_CreateFocusEnd(void);

/*
 * Factor_Match
 *
 * Attempts to match the given rule Factor, and its continuation, against
 * the current state of the pattern match.  In the case of optional and
 * repeating factors, matches the smallest number of iterations possible.
 */
extern void 
Factor_Match(Factor factor, Match_Facts facts);

/*
 * Factor_Push
 * 
 * Pushes the factor onto the beginning of the pattern matching agenda.
 */
extern void
Factor_Push(Factor factor, Match_Facts facts);

/*
 * Factor_InsertAfter
 * 
 * Inserts the factor into the string as part of a structural change.
 * The Match_Facts and deletedSegment are passed in since an AVM in
 * the change is only relative to the deleted segment or other elements
 * it is linked to by indexes or variables.
 */
extern DList_Cons
Factor_InsertAfter(Factor factor,
		   Terminal_String string,
		   DList_Cons afterCons,
		   Match_Facts facts,
		   Segment deletedSegment);

#endif  /* end of code */

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
