/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/item.c
* Description:  Implements item.h
* Author:       Brett Kessler
* Created:      Wed Apr 15 20:51:48 1992
* Modified:     Sun May  3 19:35:51 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "dlist.h"
#include "header.h"
#include "item.h"
#include "Malloc.h"
#include "match.h"
#include "segments.h"
#include <stdarg.h>

Item
Item_Create(int nFactors, ...)
{
  Item item;
  va_list args;
  int i;
  va_start(args, nFactors);
  item = Malloc_Struct(item);
  item->nFactors = nFactors;
  item->factors = Malloc(nFactors * sizeof (Factor));
  for (i = 0; i < nFactors; i++)
    item->factors[i] = va_arg(args, Factor);
  va_end(args);
  return item;
}

/* Pushes the factors in an item onto the pattern match continuation agenda.
 * We go backwards to ensure the correct sequencing, which is significant.
 */
static void
PushItem(Item item, Match_Facts facts)
{
  int f;
  for (f = item->nFactors - 1; f >= 0; f--) 
    Factor_Push(item->factors[f], facts);
}

void
Item_Match(Item item, Match_Facts facts)
{
  Cons originalCont;
  originalCont = facts->in.continuation;
  PushItem(item, facts);
  Match_Continue(facts);
  facts->in.continuation = originalCont;
}

void
Item_MatchOptional(Item item, Match_Facts facts)
{
  Match_Continue(facts);
  if (facts->out.success) return;
  Item_Match(item, facts);
}

void
Item_InsertAfter(Item new,
		 Terminal_String string,
		 DList_Cons afterCons,
		 Match_Facts facts,
		 Segment deletedSegment)
{
  int f;
  // fprintf(stderr, "Item_InsertAfter, afterCons=%X\n", afterCons);
  for (f = 0; f < new->nFactors; f++) {
    // fprintf(stderr, "  factor %d\n", f);
    afterCons =
      Factor_InsertAfter(new->factors[f],
			 string,
			 afterCons,
			 facts,
			 deletedSegment);
  }
  // fprintf(stderr, "Item_InsertAfter returns\n");
}

/* delimiter is used as a delimiter when doing resyllabification.  When
 * first needed, an utterance boundary (UB) terminal will be created
 * and placed within a DList_Cons here.  During resyllabification, this
 * delimiter is placed after the last segment in the onset, so that we
 * can do a full-span pattern match with the ONSETS pattern (which itself
 * has been extended to end in a UB match.
 */
static DList_Cons
delimiter = NULL;

/*
 * Returns TRUE iff the cons has a -syll segment in it.
 */
static Boolean
IsConsonant(DList_Cons cons)
{
  Terminal terminal;
  Segment segment;
  terminal = cons->datum;
  if (terminal->type != Terminal_segmentType) return FALSE;
  segment = terminal->u.segment;
  return !Segments_IsSyll(segment);
}

/*
 * Puts a delimiter after right-hand span of consonants.
 */
static void
DelimitOnset(DList_Cons cons)
{
   while (1) {
    cons = cons->next;
    if (!IsConsonant(cons)) {
      if (delimiter == NULL) {
	Terminal terminal;
	delimiter = Malloc_Struct(DList_cons);
	delimiter->datum = terminal = Malloc_Struct(terminal);
	terminal->type = Terminal_utteranceBoundary;
      }
      delimiter->prev = cons->prev;
      delimiter->next = cons;
      cons->prev->next = delimiter;
      cons->prev = delimiter;
      return;
    }
  }
}

/* Interchanges this element with the following. */
static void
Transpose(DList_Cons mover)
{
  DList_Cons pivot;
  DList_Cons pre;
  DList_Cons post;
  pivot = mover->next;
  pre = mover->prev;
  post = pivot->next;
  pre->next = pivot;
  pivot->prev = pre;
  pivot->next = mover;
  mover->prev = pivot;
  mover->next = post;
  post->prev = mover;
}

/* Move SB to the left of all immediately preceding consonants: i.e.,
   shift the preceding coda into the onset of this syllable. */
static void
EmptyCoda(DList_Cons sbCons)
{
  Terminal next;
  next = sbCons->next->next->datum;
  if (sbCons->next == delimiter && next->type == Terminal_utteranceBoundary) 
    return;
  while (IsConsonant(sbCons->prev)) Transpose(sbCons->prev);
}

/* return T if the SB is followed immediately by the UB delimiter */
static Boolean
IsOnsetEmpty(DList_Cons sbCons)
{
  return sbCons->next == delimiter;
}

/* Return T if the pattern is found anchored at this cons */
static Boolean
IsOnsetLegal(Item onsetPattern, DList_Cons sbCons)
{
  static struct Match_facts facts;
  facts.in.text = sbCons;
  facts.in.continuation = NULL;
  Sign_InitVarsVector(facts.in.vars);
  facts.out.success = FALSE;
  if (facts.consResource == NULL) 
    facts.consResource = List_CreateCons(500);
  else Resource_Free(facts.consResource);
  Item_Match(onsetPattern, &facts);
  return facts.out.success;
}

/* Delete the delimiter, found somewhere to the right of this cons */
static void
DeleteOnsetDelimiter(DList_Cons sbCons)
{
  DList_Cons cons;
  for (cons = sbCons; cons != delimiter; cons = cons->next);
  cons->prev->next = cons->next;
  cons->next->prev = cons->prev;
}

/* Transpose this cons with following, moving a consonant from the
   following onset to the preceding coda. */
static void
Shift1FromOnset(Terminal_String string, DList_Cons sbCons)
{
  Terminal preceding;
  preceding = sbCons->prev->datum;
  if (preceding->type == Terminal_utteranceBoundary)
    Terminal_Delete(string, sbCons->next);
  else Transpose(sbCons);
}

/* Resyllabify at this syllable boundary. */
static void
MaximizeOnset(Item onsetPattern, Terminal_String data, DList_Cons sbCons)
{
  DelimitOnset(sbCons);
  EmptyCoda(sbCons);
  while (1) {
    if (IsOnsetEmpty(sbCons) ||
	IsOnsetLegal(onsetPattern, sbCons)) {
      DeleteOnsetDelimiter(sbCons);
      return;
    }
    Shift1FromOnset(data, sbCons);
  }
}

void
Item_MaximizeOnsets(Item onsetPattern, Terminal_String data)
{
  DList_Cons terminalCons;
  for (terminalCons = data->terminals->head; 
       terminalCons != NULL; 
       terminalCons = terminalCons->next) {
    Terminal terminal = terminalCons->datum;
    if (terminal->type == Terminal_syllableBoundary)
      MaximizeOnset(onsetPattern, data, terminalCons);
  }
}

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
