/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/match.h
* Description:  Pattern matching.
* Author:       Brett Kessler
* Created:      Sun Apr 19 19:30:49 1992
* Modified:     Mon May  4 08:24:57 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#if !defined(MATCH_H)
#define MATCH_H

#include "dlist.h"
#include "header.h"
#include "list.h"
#include "rb.h"
#include "resource.h"
#include "sign.h"
#include "terminal.h"

#define Match_NINDEXES 26

/*
 * Match_IndexVector
 *
 * Records the segments encountered so far during this match for each
 * of the possible indices.
 */
typedef Segment 
Match_IndexVector[Match_NINDEXES];

/*
 * Match_In
 *
 *   The data needed to pursue pattern matching further.
 *   text is the current position in the phonetic string being matched.
 *   continuation is the ordered list of Factors that still need to be
 * matched.  Whenever a local pattern match succeeds, the function
 * attempts to match the continuation.  By leaving this responsibility
 * up to local factor matches, factors with built-in alternatives
 * (such as optionals and repeaters) can decide whether to try 
 * alternatives based on whether the rest of the match succeeds.
 *   indexes records the segments that were encountered when each
 * indexedAVM matched.  This allows the segment to be accessed when other
 * indexedAVM's have the same index.  These values are erased when 
 * an indexedAVM finds its continuation fails, so that other alternatives
 * can proceed with a clean slate.
 *   vars similarly records the values that were encountered on matching
 * alpha variables.  This is normally the value + or - for any particular 
 * variable, but for composite values will be an entire set of feature value 
 * pairs.
 */
typedef struct Match_in {
  DList_Cons text;
  Cons /* of Factor */ continuation;
  Match_IndexVector indexes; /* index char -> Segment */
  Sign_VarsVector vars;    /* of Sign_Variable */
}* Match_In;

/*
 * Match_Out
 *
 * Return data that is of importance to the caller but not to further matching.
 *  success is set to TRUE when the pattern is successfully matched to its
 * end: that is, Match_Continue finds nothing more on its continuation.
 * All intermediate functions immediately return when they see this return
 * value.
 *  end gives the cons of the the last segment that was matched.
 *  focusStart and focusEnd record where in the string we were when those
 * tokens were encountered, delimiting the part of the structural description
 * that is part of the source, i.e., the part to be changed.
 */
typedef struct Match_out {
  Boolean success;
  DList_Cons end; 
  DList_Cons focusStart;
  DList_Cons focusEnd;
}* Match_Out;

/*
 * Match_Facts
 *
 * All the state needed in pattern matching.  The consResource is used
 * when pushing items onto the continuation, so that these can all be
 * quickly reclaimed between pattern matches.
 */
typedef struct Match_facts {
  struct Match_in in;
  struct Match_out out;
  Resource consResource;
}* Match_Facts;

/*
 * Match_Continue
 * 
 * When a local matching function succeeds, it calls this function.
 * This pops an item off of the continuation agenda and tries to match
 * it.  If nothing is on the agenda, it sets out.success and out.end,
 * and it and all callers immediately return with a successful match.
 */
extern void
Match_Continue(Match_Facts facts);

#endif  /* end of code */

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
