/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/plainSignedFeature.c
* Description:  Implements plainSignedFeature.h
* Author:       Brett Kessler
* Created:      Wed Apr 15 20:25:04 1992
* Modified:     Sun Sep 27 11:19:52 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "plainSignedFeature.h"
#include "features.h"
#include "Malloc.h"
#include <stdio.h>
#include <stdlib.h>

PlainSignedFeature
PlainSignedFeature_Create(Sign sign, int featureNumber)
{
  PlainSignedFeature psf;
  psf = Malloc_Struct(plainSignedFeature);
  psf->sign = sign;
  psf->featureNumber = featureNumber;
  return psf;
}

static Boolean
SegmentHasFeature(int feature, Segment segment);

/* TRUE if any of the child features of the composite feature is found
 * in the segment. */
static Boolean
SegmentHasSuperfeature(int super, Segment segment)
{
  int c;
  for (c = 0; c < Features_COUNT; c++)
    if (Features_GetParent(c) == super && SegmentHasFeature(c, segment))
      return TRUE;
  return FALSE;
}

/* TRUE if the feature is found in this segment. */
static Boolean
SegmentHasFeature(int feature, Segment segment)
{
  unsigned long int answer;
  unsigned long int mask;
  mask = Features_GetMask(feature);
  if (!mask) return SegmentHasSuperfeature(feature, segment);
  answer = segment & mask;
  return answer != 0;
}

/* TRUE if the value of the feature inthis segment matches the constant +/- */
static Boolean
MatchConstant(Sign_Direction signDirection, int featureNumber, Segment segment)
{
  return 
    signDirection ==
      ((SegmentHasFeature(featureNumber, segment)) ? Sign_PLUS : Sign_MINUS);
}

/* Sets an alpha variable record for a simple base feature. */
static void
SetBasicFeatVar(Segment segment,
		Sign_VarFacts facts,
		Segment mask)
{
  facts->type = Sign_directionVarFactsType;
  facts->u.direction = (segment & mask) ? Sign_PLUS : Sign_MINUS;
}

/* Records alpha variable information if the variable is first encountered
 * on a composite feature.  The value of all child features is saved. 
 */
static void
SetCVar(int featNum, Segment segment, Sign_FeatVector signs)
{
  Segment mask;
  int f;
  mask = Features_GetMask(featNum);
  if (mask) signs[featNum] = (segment & mask) ? Sign_PLUS : Sign_MINUS;
  else for (f = 0; f < Features_COUNT; f++) {
    if (Features_GetParent(f) == featNum)
      SetCVar(f, segment, signs);
  }
}

static void
SetCompositeFeatVar(int parentNumber, Segment segment, Sign_VarFacts facts)
{
  int f;
  facts->type = Sign_compositeVarFactsType;
  for (f = 0; f < Features_COUNT; f++) facts->u.signs[f] = Sign_NO_DIRECTION;
  SetCVar(parentNumber, segment, facts->u.signs);
}

/* Called when a variable is encountered for the first time in a pattern
 * match.  Records the value of the feature it is attached to.
 */
static void
SetVar(int featureNumber, Segment segment, Sign_VarFacts facts)
{
  Segment mask;
  mask = Features_GetMask(featureNumber);
  if (mask)
    SetBasicFeatVar(segment, facts, mask);
  else SetCompositeFeatVar(featureNumber, segment, facts);
}

static Boolean
VerifySubfeatures(Segment segment, Sign_VarFacts facts)
{
  int f;
  /* for each feature, signs must either match, or previous variable must have
     not made an assignment for the feature. */
  for (f = 0; f < Features_COUNT; f++)
    if (facts->u.signs[f] != Sign_NO_DIRECTION && 
	!MatchConstant(facts->u.signs[f], f, segment))
      return FALSE;
  return TRUE;
}

static Boolean
VerifyVarDirection(int featureNumber, 
		   Segment segment,
		   Sign_Direction direction) __attribute__ ((noreturn));

static Boolean
VerifyVarDirection(int featureNumber, 
		   Segment segment,
		   Sign_Direction direction)
{
  fprintf(stderr, 
    "{VerifyVarDirection} featureNumber %d, segment %d, direction %d\n", 
    featureNumber, (int) segment, direction);
  exit(1);
}

/* Called when a variable is encountered in a pattern match. */
static Boolean
MatchVar(int featureNumber, Segment segment, Sign_VarFacts facts)
{
  switch (facts->type) {
  case Sign_unsetVarFactsType:  /* var not yet encountered */
    SetVar(featureNumber, segment, facts);
    return TRUE;
  case Sign_directionVarFactsType: /* var prev'ly encountered with value +/- */
    return VerifyVarDirection(featureNumber, segment, facts->u.direction);
  case Sign_compositeVarFactsType: /* var prev'ly encountered on a composite */
    return VerifySubfeatures(segment, facts);
  default:
    (void) fprintf(stderr, "Unknown variable type: %d\n", facts->type);
    exit(1);
  }
}

Boolean
PlainSignedFeature_Match(PlainSignedFeature feature,
			 Segment segment,
			 Sign_VarsVector vars)
{
  switch (feature->sign->type) {
  case Sign_constantType:
    return 
      MatchConstant(feature->sign->u.direction,
		    feature->featureNumber, 
		    segment);
  case Sign_variableType:
    return
      MatchVar(feature->featureNumber,
	       segment,
	       &vars[feature->sign->u.variable]);
  default: 
    (void) fprintf(stderr, "Unknown type of sign: %d\n", feature->sign->type);
    exit(1);
  }
}

/* When structural change has - on a composite feature, remove all
 * its children features from the segment.
 */
static Segment
RemoveChildren(int featureNumber, Segment segment)
{
  int f;
  for (f = 0; f < Features_COUNT; f++)
    if (Features_GetParent(f) == featureNumber) {
      Segment mask;
      mask = Features_GetMask(f);
      if (mask) segment &= ~mask;
      else segment = RemoveChildren(f, segment);
    }
  return segment;
}

/* When a + or - is encountered in a structural change, add or remove
 * the feature in question.  But note one can't add a composite feature:
 * that would be ambiguous.
 */
static Segment
AddConstant(Sign_Direction signDirection, int featureNumber, Segment segment)
{
  Segment mask;
  mask = Features_GetMask(featureNumber);
  if (mask) {
    return (signDirection == Sign_PLUS) ? (segment | mask) : (segment & ~mask);
  }
  else if (signDirection == Sign_MINUS) 
    return RemoveChildren(featureNumber, segment);
  else {
    (void) 
      fprintf(stderr,
	      "Don't know what it means to add a constant composite"
	      " feature.\n");
    exit(1);
  }
}

/* When an alpha variable is encountered in a structural change but
 * was first found on a composite variable, copy the features there
 * encountered, provided the types match. */
static Segment
AddComposite(Sign_FeatVector feats, int featureNumber, Segment segment)
{
  int f;
  for (f = 0; f < Features_COUNT; f++)
    if (feats[f] != Sign_NO_DIRECTION) {
      if (!Features_IsDescendant(f, featureNumber)) {
	(void) fprintf(stderr, "Variable type mismatch.\n");
	exit(1);
      }
      segment = AddConstant(feats[f], f, segment);
    }
  return segment;
}

/* When a variable is encountered in a structural change, copy the
 * feature values found when that variable was encountered in the
 * structural description.
 */
static Segment
AddVar(int featureNumber, Segment segment, Sign_VarFacts facts)
{
  switch (facts->type) {
  case Sign_directionVarFactsType:
    return AddConstant(facts->u.direction, featureNumber, segment);
  case Sign_compositeVarFactsType:
    return AddComposite(facts->u.signs, featureNumber, segment);
  default: 
    (void) fprintf(stderr, "Attempt to add variable of type %d\n", 
		   facts->type);
    exit(1);
  }
}

Segment
PlainSignedFeature_Add(PlainSignedFeature psf, 
		       Segment segment,
		       Sign_VarsVector variables)
{
  Sign sign;
  sign = psf->sign;
  switch (sign->type) {
    case Sign_constantType:
      return AddConstant(sign->u.direction, psf->featureNumber, segment);
    case Sign_variableType:
      return
	AddVar(psf->featureNumber, segment, &variables[sign->u.variable]);
    default:
      (void) fprintf(stderr, "Unknown type of sign: %d\n", sign->type);
      exit(1);
  }  
}

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
