%{

 /* derive/readWordsYacc.y parses the input stream to read 
  * sequences of words in lexical form and an expected sandhi.
  * It calls the rule system to do the derivation and see if the
  * final result matches the expectation.
  */

#include "ruleSystem.h"
#include "segment.h"
#include "terminal.h"

static Boolean
useLaTeX = FALSE;

/* the sequence of words in underlying form, with boundaries explicitly
 * inserted into the string.
 */
static Terminal_String
inputData;

/* the sequence of phones describing the expected sandhi, as stated in
 * the input file. */
static Terminal_String
expectedOutput;

/* are we reading the input words or the expected output? */
static enum {inputHalf, outputHalf}
whichHalf = inputHalf;

/* how many times did the derivation result fail to match the expected
 * output? */
static int
nErrors = 0;

/* Bit mask representing the lexical word attributes most recently read. */
static unsigned long int
latestProperties = 0;

/* Adds the terminal (string element) to the string being composed. */
static void
AddTerminal(Terminal terminal)
{
  if (whichHalf == inputHalf) Terminal_Add(inputData, terminal);
  else Terminal_Add(expectedOutput, terminal);
}


/* Boundaries are freely inserted.  Inter-syllable and inter-word
 * boundaries are marked in the input.  In addition, we add syllable
 * boundaries inside word boundaries; word boundaries at the utterance
 * edges; and utterance boundaries around everything.
 */


static void
AddSyllableBoundary(void)
{
  AddTerminal(Terminal_CreateSyllableBoundary());
}

static void
AddWordBoundary(void)
{
  AddTerminal(Terminal_CreateWordBoundary(latestProperties));
}

static void
AddUtteranceBoundary(void)
{
  AddTerminal(Terminal_CreateUtteranceBoundary());
}

%}

%union {
  int integer;
  Terminal_Properties properties;
}

%token <integer> ATTRIBUTE_NAME
%token <integer> BECOMES
%token <integer> END_PROPS
%token <integer> PLUS
%token <integer> PROP_SEP
%token <integer> SEGMENT_NAME
%token <integer> START_PROPS
%token <integer> SYLL_SEP
%token <integer> TERMINATOR 

%type <properties> properties
%type <properties> proplist
%type <integer> property

%%

file:                  examples
                       { if (nErrors)
                           (void) printf("%d errors.\n", nErrors);
			 else (void) puts("No errors.");
		       }

examples:              example
                     | examples example
                     ;

example:               words becomes syllables TERMINATOR
                       {   
			 Boolean okay;
                         /* AddSyllableBoundary(); */
                         AddUtteranceBoundary();
			 okay =
                           RuleSystem_Run(inputData, expectedOutput, useLaTeX);
			 if (!okay) nErrors++;
                         /* Transient resources are freed between examples. */
			 Terminal_ReinitString(inputData);
			 Terminal_ReinitString(expectedOutput);
                         Terminal_Free();
			 whichHalf = inputHalf;
                         latestProperties = 0;
                         AddUtteranceBoundary();
                         AddWordBoundary();
                         AddSyllableBoundary();
		       }
                     ;

becomes:               BECOMES
                       {
                         AddSyllableBoundary();
                         AddWordBoundary();
                         AddUtteranceBoundary();
                         whichHalf = outputHalf; 
                         AddUtteranceBoundary();
                         /* AddSyllableBoundary(); */
                       }
                     ;

words:                 word
                     | words wordBoundary word
                     ;

wordBoundary:          PLUS
                       { 
			 AddSyllableBoundary(); 
			 AddWordBoundary(); 
			 AddSyllableBoundary(); 
		       }
                     ;

word:                  syllables 
                       { latestProperties = 0; }
                     | syllables properties
                       { latestProperties = $2; }
                     ;

syllables:             syllable
                     | syllables syllBoundary syllable
                     ;

syllBoundary:          SYLL_SEP
                       { AddSyllableBoundary(); }
                     ;

syllable:              segment
                     | syllable segment
                     ;

segment:               SEGMENT_NAME
                       { AddTerminal(Terminal_CreateSegment($1)); }
                     ;

properties:            START_PROPS proplist END_PROPS
                       { $$ = $2; }
                     ;

proplist:              property
                       { $$ = (1uL << $1) ; }
                     | proplist PROP_SEP property
                       { $$ = $1 | (1uL << $3); }
                     ;

property:              ATTRIBUTE_NAME
                       { $$ = $1; }
                     ;

%%


#include "readWordsLex.c"

static int
ReadWordsYacc_Position(void)
{
  return (int) ftell(yyin);
}

int
yyerror(char* s) __attribute__ ((noreturn));

int
yyerror(char* s)
{
  int position;
  position = ReadWordsYacc_Position();
  fprintf(stderr, "%s in file", s);
  if (position >= 0) fprintf(stderr, ", near position %d\n", position);
  else putchar('\n');
  exit(1);
}

extern void
ReadWordsYacc_Parse(Boolean latex)
{
  useLaTeX = latex;
  inputData = Terminal_CreateString();
  AddUtteranceBoundary();
  AddWordBoundary();
  AddSyllableBoundary();
  expectedOutput = Terminal_CreateString();
  yyparse();
}
