/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/rule.c
* Description:  Implements rule.h
* Author:       Brett Kessler
* Created:      Wed Apr 15 21:34:53 1992 (Brett Kessler)
* Modified:     Sun May  3 20:10:21 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "rule.h"
#include "Malloc.h"
#include "match.h"
#include <string.h>

char*
Rule_longestName = NULL;

int
Rule_longestNameLen = 0;

Rule
Rule_CreateOrdinary(char* name,
		    Item original,
		    Item change)
{
  Rule r;
  r = Malloc_Struct(rule);
  r->type = Rule_ordinaryType;
  r->name = name;
  r->original = original;
  r->change = change;
  return r;
}

Rule
Rule_CreateSpecial(Rule_Type type)
{
  Rule r;
  r = Malloc_Struct(rule);
  r->type = type;
  return r;
}

char*
Rule_Name(Rule rule)
{
  switch (rule->type) {
  case Rule_ordinaryType: return rule->name;
  case Rule_maximizeOnsets: return "Maximize Onsets";
  default: return "";
  }
}

int
Rule_NameLength(Rule rule)
{
  char* name;
  name = Rule_Name(rule);
  return strlen(name);
}

/* When a structural description matches, this is called.  "new" is the
 * structural change pattern.  "facts" tells where the source element
 * is that is to be changed, along with the values assigned to variables
 * and indices.
 */
static void
Replace(Terminal_String string, Match_Facts facts, Item new)
{
  DList_Cons prev;
  Segment deletedSegment;
  Boolean zeroMatch;
  zeroMatch = facts->out.focusStart->prev == facts->out.focusEnd;
  /* Find the matched segment, since this is what unindexed AVMs in the
     structural change implicitly refer to.  This has no value if
     more than one segment is matched. */
  deletedSegment =
    (zeroMatch) ? 0 :
      Terminal_ExtractSegment(facts->out.focusStart, facts->out.focusEnd);
  prev = facts->out.focusStart->prev;
  /* Excise the matched segment(s), and replace with the new segment */
  if (!zeroMatch)
    Terminal_Remove(string,
		    facts->out.focusStart,
		    facts->out.focusEnd);
  Item_InsertAfter(new, string, prev, facts, deletedSegment);
}

/* Runs a rule other than MaximizeOnsets.  Initializes variables and
 * resources, then tries to do a pattern match at the current 
 * position.  If this succeeds, calls Replace to make the structural
 * change.
 */
static void
RunOrdinaryRule(Rule rule, Terminal_String string, DList_Cons text)
{
  static struct Match_facts facts;
  int s;
  //fprintf(stderr, "RunOrdinaryRule(%s)\n", rule->name);
  facts.in.text = text;
  if (facts.consResource == NULL) facts.consResource = List_CreateCons(500);
  else Resource_Free(facts.consResource);
  facts.in.continuation = NULL;
  Sign_InitVarsVector(facts.in.vars);
  for (s = 0; s < 26; s++) facts.in.indexes[s] = 0;
  facts.out.focusStart = NULL;
  facts.out.success = FALSE;
  Item_Match(rule->original, &facts);
  //fprintf(stderr, "  success = %d\n", facts.out.success);
  if (!facts.out.success) return;
  Replace(string, &facts, rule->change);
}

/* Runs the rule on the input string.  Tries it at every position
 * in the string, working from left to right, regardless of whether
 * there are any matches.
 */
static void
RunRule(Rule rule, Terminal_String data, Item onsetPattern)
{
  // fprintf(stderr, "RunRule(%s)\n", rule->name);
  switch (rule->type) {
  case Rule_ordinaryType: {
    DList_Cons terminalCons;
    for (terminalCons = data->terminals->head; 
       terminalCons != NULL; 
       terminalCons = terminalCons->next) {
      RunOrdinaryRule(rule, data, terminalCons);
    }
    break;
  }
  case Rule_maximizeOnsets: 
    Item_MaximizeOnsets(onsetPattern, data); break;
  }
  // fprintf(stderr, "RunRule returns\n");
}

/* Runs a rule against the input string.  If there results any change
 * in that string, prints out a line naming the rule and showing the
 * changed state.  Null applications don't count.
 */
void
Rule_Run(Rule rule, Terminal_String data, Item onsetPattern, Boolean latex)
{
  static Terminal_String original = NULL;
  if (original == NULL) original = Terminal_CreateString();
  else Terminal_ReinitString(original);
  Terminal_CopyString(original, data);
  RunRule(rule, data, onsetPattern);
  if (!Terminal_EqualString(original, data)) //segmentation violation
  {
    int n;
    (void) fputs(Rule_Name(rule), stdout);
    (void) printf(":");
    if (latex) (void) fputs("&", stdout);
    for (n = Rule_NameLength(rule); n < Rule_longestNameLen + 1; n++)
      putchar(' ');
    Terminal_PrintString(data, stdout, latex);
    if (latex) (void) fputs(" \\\\", stdout);
    putchar('\n');
  }
}

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
