/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/terminal.c
* Description:  Implements terminal.h
* Author:       Brett Kessler
* Created:      Sun Apr 19 11:11:04 1992
* Modified:     Sat May 23 09:31:25 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "terminal.h"
#include "dlist.h"
#include "Malloc.h"
#include "resource.h"
#include "segments.h"
#include "../frag/attributesH"
#include "../frag/attributesC"
#include <stdlib.h>

/* Terminals are allocated from this resource, which is cleared between
 * examples.
 */
static Resource
terminalResource = NULL;

void
Terminal_Add(Terminal_String string, Terminal terminal)
{
  DList_Add(string->terminals, terminal);
}

Terminal_String
Terminal_CreateString()
{
  Terminal_String string;
  string = Malloc_Struct(Terminal_string);
  string->terminals = DList_Create();
  string->terminals->consResource = DList_CreateCons(500);
  return string;
}

void
Terminal_ReinitString(Terminal_String string)
{
  Resource_Free(string->terminals->consResource);
  string->terminals->count = 0;
  string->terminals->head = string->terminals->tail = NULL;
}

static Terminal
GetTerminal(void)
{
  if (terminalResource == NULL) 
    terminalResource = Resource_Create(sizeof (struct terminal), 500);
  return Resource_Get(terminalResource);
}

Terminal
Terminal_CreateWordBoundary(Terminal_Properties props)
{
  Terminal t;
  t = GetTerminal();
  t->type = Terminal_wordBoundary;
  t->u.properties = props;
  return t;
}

Terminal
Terminal_CreateSyllableBoundary(void)
{
  Terminal t;
  t = GetTerminal();
  t->type = Terminal_syllableBoundary;
  return t;
}

Terminal
Terminal_CreateUtteranceBoundary(void)
{
  Terminal t;
  t = GetTerminal();
  t->type = Terminal_utteranceBoundary;
  return t;
}

Terminal
Terminal_CreateSegment(int id)
{
  Terminal t;
  t = GetTerminal();
  t->type = Terminal_segmentType;
  t->u.segment = Segments_SegmentFromID(id);
  return t;
}

/* Create a segment Terminal for a given segment bit mask (not an ID)
 */
static Terminal
CreateSegmentItself(Segment segment)
{
  Terminal t;
  t = GetTerminal();
  t->type = Terminal_segmentType;
  t->u.segment = segment;
  return t;
}

void
Terminal_Free(void)
{
  Resource_Free(terminalResource);
}

static Terminal
Copy(Terminal old) {
  Terminal new = GetTerminal();
  new->type = old->type;
  switch (new->type) {
  case Terminal_segmentType: 
    new->u.segment = old->u.segment;
    break;
  default:
    new->u.properties = old->u.properties;
  }
  return new;
}

void
Terminal_CopyString(Terminal_String new, Terminal_String old)
{
  DList_Cons terminalCons;
  for (terminalCons = old->terminals->head; 
       terminalCons != NULL; 
       terminalCons = terminalCons->next) {
    Terminal terminal = terminalCons->datum;
    Terminal_Add(new, Copy(terminal)); //added 20040724
  }
}

/* TRUE iff the two Terminal's refer to the same value. */
static Boolean
Terminal_Equal(void* t1_arg, void* t2_arg)
{
  Terminal t1 = (Terminal) t1_arg;
  Terminal t2 = (Terminal) t2_arg;
  if (t1 == NULL || t2 == NULL) {
    (void) fprintf(stderr, "Terminal_Equal(%X, %X)\n", 
		   (unsigned int) t1, (unsigned int) t2);
    exit(1);
  }
  if (t1->type != t2->type) return FALSE;
  switch (t1->type) {
  case Terminal_segmentType: return t1->u.segment == t2->u.segment;
  case Terminal_wordBoundary: return t1->u.properties == t2->u.properties;
  default: return TRUE;
  }
}

Boolean
Terminal_EqualString(Terminal_String s1, Terminal_String s2)
{
  return DList_Equal(s1->terminals, s2->terminals, Terminal_Equal);
}

void
Terminal_PrintString(Terminal_String string, FILE* file, Boolean latex)
{
  DList_Cons terminalCons;
  if (string == NULL) return;
  for (terminalCons = string->terminals->head; 
       terminalCons != NULL; 
       terminalCons = terminalCons->next) {
    Terminal terminal = terminalCons->datum;
    Terminal_Print(terminal, file, latex);
  }
}

void
Terminal_Delete(Terminal_String string, DList_Cons cons)
{
  DList list;
  list = string->terminals;
  if (cons->prev == NULL) list->head = cons->next;
  else cons->prev->next = cons->next;
  if (cons->next == NULL) list->tail = cons->prev;
  else cons->next->prev = cons->prev;
  list->count--;
}

/* Prints word attributes by name, within square brackets. */
static void
PrintProperties(Terminal_Properties props, FILE* file)
{
  int attributeNumber;
  Boolean alreadyPrintedOne = FALSE;
  if (!props) return;
  (void) putc('[', file);
  for (attributeNumber = 0; 
       attributeNumber < Attributes_COUNT;
       attributeNumber++)
    if (props & (1uL << attributeNumber)) {
      if (alreadyPrintedOne) (void) fputs(", ", file);
      (void) fputs(Attributes_Names[attributeNumber], file);
      alreadyPrintedOne = TRUE;
    }
  (void) putc(']', file);
}

void
Terminal_Print(Terminal terminal, FILE* file, Boolean latex)
{
  switch(terminal->type) {
  case Terminal_segmentType:
    Segments_PrintSegment(terminal->u.segment, file, latex); 
    break;
  case Terminal_syllableBoundary: 
    (void) putc('-', file);
    break;
  case Terminal_wordBoundary: 
    if (terminal->u.properties)
      PrintProperties(terminal->u.properties, file);
    (void) putc(' ', file); 
    break;
  case Terminal_utteranceBoundary: 
    if (latex) (void) printf("\\");
    (void) putc('#', file); 
    break;
  }
}

void
Terminal_Remove(Terminal_String string, DList_Cons first, DList_Cons last)
{
  DList_Cons cons;
  for (cons = first; cons != NULL; cons = cons->next) {
    Terminal_Delete(string, cons);
    if (cons == last) break;
  }
}

Segment
Terminal_ExtractSegment(DList_Cons first, DList_Cons last)
{
  Segment segment = 0;
  DList_Cons cons;
  for (cons = first; cons->prev != last; cons = cons->next) {
    Terminal terminal;
    terminal = cons->datum;
    if (terminal->type == Terminal_segmentType) {
      if (segment) return 0;
      segment = terminal->u.segment;
    }
  }   
  return segment;
}

static DList_Cons
InsertTerminal(Terminal_String string, Terminal terminal, DList_Cons afterCons)
{
  DList_Cons cons;
  DList list;
  // fprintf(stderr, "InsertTerminal(string = %X, afterCons = %X)\n", 
  // string, afterCons);
  list = string->terminals;
  // fprintf(stderr, "InsertTerminal list = %X\n", list);
  cons = Resource_Get(string->terminals->consResource);
  // fprintf(stderr, "InsertTerminal cons = %X\n", cons);
  cons->datum = terminal;
  cons->prev = afterCons;
  if (afterCons == NULL) {
    cons->next = list->head;
    list->head = cons;
  } else {
    // fprintf(stderr, "InsertTerminal, afterCons=%X\n", afterCons);
    cons->next = afterCons->next;
    afterCons->next = cons;
  }
  // fprintf(stderr, "InsertTerminal B, cons->next = %X\n", cons->next);
  if (cons->next == NULL) list->tail = cons;
  else cons->next->prev = cons;
  // fprintf(stderr, "InsertTerminal C\n");
  list->count++;
  // fprintf(stderr, "InsertTerminal returns\n");
  return cons;
}

DList_Cons
Terminal_InsertSegment(Terminal_String string,
		       Segment segment,
		       DList_Cons afterCons)
{
  Terminal terminal;
  terminal = CreateSegmentItself(segment);
  return InsertTerminal(string, terminal, afterCons);
}

DList_Cons
Terminal_InsertBoundary(Terminal_String string,
			Terminal_Type boundary,
			DList_Cons afterCons)
{
  Terminal terminal;
  // fprintf(stderr, "Terminal_InsertBoundary\n");
  switch (boundary) {
  case Terminal_syllableBoundary:
    // fprintf(stderr, "Terminal_InsertBoundary syllableBoundary\n"); 
    terminal = Terminal_CreateSyllableBoundary();
    break;
  case Terminal_utteranceBoundary:
    terminal = Terminal_CreateUtteranceBoundary();
    break;
  default:
    (void) fprintf(stderr, "%d is not a recognized boundary type.\n", 
		   (int) boundary);
    exit(1);
  }
  // fprintf(stderr, "Terminal_InsertBoundary A\n");
  DList_Cons new = InsertTerminal(string, terminal, afterCons);
  // fprintf(stderr, "Terminal_InsertBoundary returns %X\n", new);
  return new;
}

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
