/*  -*- mode:     C -*- --------------------------------------------------
* File:         derive/terminal.h
* Description:  An element in the text string (segment or boundary).
* Author:       Brett Kessler
* Created:      Sun Apr 19 09:55:13 1992
* Modified:     Sat May 23 09:26:12 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#if !defined(TERMINAL_H)
#define TERMINAL_H

#include "dlist.h"
#include "segment.h"
#include <stdio.h>

typedef enum Terminal_type {
  Terminal_segmentType,
  Terminal_syllableBoundary,
  Terminal_wordBoundary,
  Terminal_utteranceBoundary
} Terminal_Type;

/*
 * Terminal_Properties
 *
 * The lexical attributes of a word, as a bit mask.
 */
typedef unsigned long
Terminal_Properties;

/*
 * Terminal
 *
 * A terminal in a text string.
 */
typedef struct terminal {
  Terminal_Type type;
  union {
    Segment segment; /* Terminal_segmentType */
    Terminal_Properties properties; /* Terminal_wordBoundary */
  } u;
}* Terminal;

/*
 * Terminal_String
 *
 * A text string: a sequence of terminals, implemented as a doubly-linked
 * List for ease of insertion and deletion.
 */
typedef struct Terminal_string {
  DList terminals;
}* Terminal_String;

/*
 * Terminal_CreateString
 *
 * Creates an empty Terminal_String.
 */
extern Terminal_String
Terminal_CreateString(void);

/*
 * Terminal_Add
 *
 * Adds a Terminal to the Terminal_String.
 */
extern void
Terminal_Add(Terminal_String string, Terminal terminal);

/*
 * Terminal_CreateWordBoundary
 *
 * Creates a Terminal representing a word boundary following a word
 * with the given lexical attributes.
 */
extern Terminal
Terminal_CreateWordBoundary(Terminal_Properties props);

/*
 * Terminal_CreateSyllableBoundary
 */
extern Terminal
Terminal_CreateSyllableBoundary(void);

/*
 * Terminal_CreateUtteranceBoundary
 */
extern Terminal
Terminal_CreateUtteranceBoundary(void);

/*
 * Terminal_CreateSegment
 *
 * Creates a segment Terminal for the segment with the given numeric ID
 * (not feature mask)
 */
extern Terminal
Terminal_CreateSegment(int id);

/*
 * Terminal_ReinitString
 *
 * Empties the string, reclaiming all the cons cells for reuse.
 */
extern void
Terminal_ReinitString(Terminal_String string);

/*
 * Terminal_Free
 *
 * Deletes all Terminals, reclaiming their memory for reuse.
 */
extern void
Terminal_Free(void);

/*
 * Terminal_CopyString
 *
 * Copies the 2d string into the 1st.
 */
extern void
Terminal_CopyString(Terminal_String new, Terminal_String old);

/*
 * Terminal_EqualString
 *
 * TRUE iff the two strings have the same terminals in the same order.
 */
extern Boolean
Terminal_EqualString(Terminal_String s1, Terminal_String s2);

/*
 * Terminal_PrintString
 * 
 * Displays the string on the stream.  Chooses ASCII or LaTeX representations
 * depending on the Boolean.
 */
extern void
Terminal_PrintString(Terminal_String string, FILE* file, Boolean latex);

/* 
 * Terminal_Print
 *
 * Prints a Terminal on the stream, in the format selected by the Boolean.
 */
extern void
Terminal_Print(Terminal terminal, FILE* file, Boolean latex);

/*
 * Terminal_Delete
 *
 * Deletes the terminal from the string.
 */
extern void
Terminal_Delete(Terminal_String string, DList_Cons cons);

/*
 * Terminal_Remove
 *
 * Removes from the string all the terminals from first to last,
 * inclusive.
 */
extern void
Terminal_Remove(Terminal_String string, DList_Cons first, DList_Cons last);

/*
 * Terminal_ExtractSegment
 *
 * Sees if a single segment occurs in the span from first to last,
 * inclusive (that is, boundaries are ignored). If so, returns it.
 * If there were more or fewer than one, returns 0.  (For use in the
 * structural change; an unindexed AVM has meaning there only if we
 * had matched in the source exactly one segment).
 */
extern Segment
Terminal_ExtractSegment(DList_Cons first, DList_Cons last);

/*
 * Terminal_InsertSegment
 *
 * Inserts one segment into the string.
 */
extern DList_Cons
Terminal_InsertSegment(Terminal_String string,
		       Segment segment,
		       DList_Cons afterCons);

/*
 * Terminal_InsertBoundary
 *
 * Inserts a boundary into the string.
 */
extern DList_Cons
Terminal_InsertBoundary(Terminal_String string,
			Terminal_Type boundary,
			DList_Cons afterCons);

#endif  /* end of code */

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
