/*  -*- mode:     C -*- --------------------------------------------------
* File:         features/avm.c
* Description:  Implements avm.h.  Builds AVM for FEATURES, prints C files.
* Author:       Brett Kessler
* Created:      Sat Mar 28 18:48:41 1992
* Modified:     Sat May  2 14:37:59 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "avm.h"
#include "fvPair.h"
#include "list.h"
#include "Malloc.h"
#include <stdio.h>

AVM
AVM_Create(List fvPairs)
{
  AVM avm;
  avm = Malloc_Struct(avm);
  avm->fvPairs = fvPairs;
  return avm;
}

void
AVM_Print(AVM avm, FILE* file, int indent)
{
  Cons fvPairCons;
  FvPair fvPair;
  (void) fprintf(file, "[\n");
  List_ForEach(avm->fvPairs, fvPairCons, fvPair) {
    int i;
    if (fvPairCons != avm->fvPairs->head) (void) fprintf(file, ";\n");
    for (i = 0; i < indent; i++) (void) fprintf(file, " ");
    FvPair_Print(fvPair, file, indent);
  }
  (void) fprintf(file, "]");
}

/*
 * Assign a sequential numerical id (starting from 0) to each feature,
 * in the order the features were encountered in the input file (the exact
 * order doesn't matter, just so long as we are consistent within this
 * file).
 */
static void
EnumerateFeatures(AVM avm, int* count)
{
  Cons fvPairCons;
  FvPair fvPair;
   List_ForEach(avm->fvPairs, fvPairCons, fvPair) {
    ValueType vt;
    fvPair->id = (*count)++;
    vt = fvPair->valueType;
    if (vt->type == ValueType_avm)
      EnumerateFeatures(vt->avm, count);
  }    
}

/* 
 * Print the Lex file code (featuresLex): code fragment will recognize the
 * name of a feature (within SEGMENTS or RULES file), return the identifier
 * token FEATURE_NAME, and set yylval to the numerical ID we have assigned the
 * feature,
 */
static void
PrintFeaturesLex(AVM avm, FILE* file)
{
  Cons fvPairCons;
  FvPair fvPair;
   List_ForEach(avm->fvPairs, fvPairCons, fvPair) {
    ValueType vt;
    (void)
      fprintf(file,
	      "\"%s\"\t\t{ yylval.integer = %d; return FEATURE_NAME; }\n",
	      fvPair->featName, fvPair->id);
    vt = fvPair->valueType;
    if (vt->type == ValueType_avm)
      PrintFeaturesLex(vt->avm, file);
  }    
}

/* 
 * Generates C include file featuresH.
 */
static void
PrintFeaturesH(FILE* file, int nFeatures)
{
  (void) fprintf(file, "#define Features_COUNT %d\n\n", nFeatures);
}

/*
 * In the C file (featuresC), prints the body of the Features_Names
 * definition, an array telling the name of each feature, given its ID.
 */
static void
PrintFeaturesCInner(AVM avm, FILE* file)
{
  Cons fvPairCons;
  FvPair fvPair;
  List_ForEach(avm->fvPairs, fvPairCons, fvPair) {
    ValueType vt;
    (void)
      fprintf(file,
	      "  \"%s\",  /* %d */\n",
	      fvPair->featName,
	      fvPair->id);
    vt = fvPair->valueType;
    if (vt->type == ValueType_avm)
      PrintFeaturesCInner(vt->avm, file);
  }    
}

/*
 * In the C file featuresC, prints the body of the Features_parents
 * array, which tells the composite feature that a feature is immediately
 * dominated by (e.g., the parent of ant is Coronal; of Coronal, Place).
 * Top-level features (e.g., cons, Place) get the value Features_NONE.
 */
static void
PrintFeaturesParents(AVM avm, FILE* file, FvPair parent)
{
  Cons fvPairCons;
  FvPair fvPair;
  List_ForEach(avm->fvPairs, fvPairCons, fvPair) {
    ValueType vt;
    if (parent == NULL)
      (void) fprintf(file,
		     "  Features_NONE, /* %s is top-level */\n",
		     fvPair->featName);
    else
      (void)
	fprintf(file,
		"  %d,  /* parent of %s is %s */\n",
		parent->id,
		fvPair->featName,
		parent->featName);
    vt = fvPair->valueType;
    if (vt->type == ValueType_avm)
      PrintFeaturesParents(vt->avm, file, fvPair);
  }    
}

/*
 * In the C file featuresC, prints the body of the Features_mask array,
 * which assigns for each feature ID a bit string to be used as a mask
 * for representing its value in a packed representation.  For composite
 * features, it is 0.
 */
static void
PrintFeaturesMask(AVM avm, FILE* file, unsigned long int* mask)
{
  Cons fvPairCons;
  FvPair fvPair;
  List_ForEach(avm->fvPairs, fvPairCons, fvPair) {
    ValueType vt;
    vt = fvPair->valueType;
    (void)
      fprintf(file,
	      "/* %d */ %lu, /* %s */\n",
	      fvPair->id, 
	      (vt->type == ValueType_avm) ? 0 : *mask,
	      fvPair->featName);
    if (vt->type == ValueType_avm)
      PrintFeaturesMask(vt->avm, file, mask);
    else (*mask) *= 2;
  }    
}

/*
 * Prints the C implementation module fragment (sandhi/frags/featuresC).
 */
static void
PrintFeaturesC(AVM avm, FILE* file)
{
  unsigned long int mask = 1;
  (void) fputs("char* Features_Names[Features_COUNT + 1] = {\n", file);
  PrintFeaturesCInner(avm, file);
  (void) fputs("  \"ILLEGAL\"\n"
	       "};\n\n"
	       "Features_Id Features_parent[Features_COUNT + 1] = {\n",
	       file);
  PrintFeaturesParents(avm, file, NULL);
  (void) fputs("  Features_NONE\n"
	       "};\n\n"
	       "unsigned long int Features_mask[Features_COUNT + 1] = {\n",
	       file);
  PrintFeaturesMask(avm, file, &mask);
  (void) fputs("  0\n"
	       "};\n\n",
	       file);
}

/* 
 * Prints the three source code files in sandhi/frag/: featuresLex,
 * featuresH, and featuresC.
 */
void
AVM_TransduceToPhonesReader(AVM avm,
			    FILE* lexFile,
			    FILE* featureIncludeFile,
			    FILE* featureCFile)
{
  int count = 0;
  EnumerateFeatures(avm, &count);
  PrintFeaturesLex(avm, lexFile);
  PrintFeaturesH(featureIncludeFile, count);
  PrintFeaturesC(avm, featureCFile);
}

/*
******************************************************************************
* Revisions:
* RCS $Log: avm.c,v $
* RCS Revision 1.1  2000/02/29 11:06:23  bkessler
* RCS Backup
* RCS
******************************************************************************

******************************************************************************
*                                End of file
******************************************************************************
*/
