/*  -*- mode: C -*- -----------------------------------------
 * File:           list.h
 * Decsription:    Singly linked list of pointers (FIFO queue).
 * Author:         Brett Kessler
 * Created:        10-Jan-92
 * Language:       C
 **************************************************************
 */

#if !defined(LIST_H)
#define LIST_H

#include "header.h"
#include "Malloc.h"
#include "resource.h"
#include <stdlib.h>

/*
 * Cons
 *
 * A list is built up from a singly linked list of Cons. A Cons contains
 * a client datum and a pointer to the next Cons.  next is NULL at end of
 * list.
 */
typedef struct cons {
  void* datum;
  struct cons* next;
}* Cons;

/*
 * List
 * A handle on a list.  count tells how many data.  head is first element
 * of list, tail is last element (facilitating adding elements to end).
 * Optional resource is used for making new Cons; if NULL, they will be
 * malloc'ed.
 */
typedef struct list {
  unsigned int count;
  Cons head;
  Cons tail;
  Resource consResource;
}* List;

/*
 * List_Enumerator
 *
 * Client callback type. First argument is a datum the client put into
 * the list; the second argument is the closure the client supplied to the
 * enumerating procedure.
 */
typedef void (*List_Enumerator)(void* datum, void* closure);

/*
 * List_Init
 *
 * If client allocates its own struct list, e.g., as an automatic variablem
 * it must call this function to initialize the structure.  NULLs resource heap
 * variable.
 */
extern void
List_Init(List list);

/*
 * List_Reinit
 *
 * Sets list to emoty and frees the resource heap, if any.
 */
extern void
List_Reinit(List list);

/*
 * List_Zero
 *
 * Sets list to empty, but doesn't touch the resource heap.
 */
extern void
List_Zero(List list);

/*
 * List_Create
 *
 * Creates a new list.  Client does not need to call List_Init.
 */
extern List
List_Create(void);

/*
 * List_Add
 *
 * Adds a new element to the end of the list.  If list is NULL, creates a
 * new one. Returns the list. If the list has a consResource, that is
 * used for the new cons; otherwise it will be malloc'ed.
 */
extern List
List_Add(List list, void* datum);

/*
 * List_AddCons
 *
 * Like List_Add, but returns the new cons.
 */
extern Cons
List_AddCons(List list, void* datum);

#define List_DEFAULT_CONS_PER_BLOCK 1000

/* 
 * List_CreateCons
 *
 * Creates a Resource that can be used for allocating Cons elements
 * (confer List_AddR and the Resource interface).
 */
extern Resource
List_CreateCons(unsigned int consPerBlock);

/*
 * List_Count
 *
 * Tells how many items are in the list. If list is NULL, returns 0.
 */
extern unsigned int
List_Count(List list);

/*
 * List_FreePtr
 *
 * Like free(3), but takes and ignores a second argument. This can be passed
 * to List_Free to handle the simplest case where disjoint data nee to be
 * freed on freeing a list.
 */
extern void
List_FreePtr(void* datum, void* closure);

/* 
 * List_Free
 *
 * Destroys the list. If callback is not NULL, it will be called once
 * for each item. (The List modules will not of itself free the client's
 * data; this is the client's opportunity to do so.)
 * If list is NULL, does nothing.
 */
extern void
List_Free(List list,
	  List_Enumerator callback,
	  void* closure);

/*
 * List_FreeListOnly
 *
 * Destroys the list, without touching the included data.
 */
extern void
List_FreeListOnly(List list);

/*
 * List_Enum
 *
 * Calls the callback once for each item in the list, from front to back.
 * That is, items entered first will be passed back first (FIFO).
 * If list is NULL, does nothing.
 */
extern void
List_Enum(List list,
	  List_Enumerator callback,
	  void* closure);

/*
 * List_ForEach
 *
 * Enumeration without callback.
 */
#define List_ForEach(list, cons, value) \
  for ((cons) = (list)->head, \
         (value) = (((cons) == NULL) ? NULL : (cons)->datum); \
       (cons) != NULL; \
       (cons) = (cons)->next, \
         (value) = (((cons) == NULL) ? NULL : (cons)->datum))

/*
 * List_First
 *
 * Returns the first datum on the list, without removing it.
 */
extern void*
List_First(List list);

/*
 * List_Last
 *
 * Returns the last datum on the list, without removing it.
 */
extern void*
List_Last(List list);

/*
 * List_Pop
 *
 * Removes and returns the first element on the list.
 */
extern void*
List_Pop(List list);

/*
 * List_Comparer
 */
typedef Boolean (*List_Comparer)(void* datum1, void* datum2);

/*
 * List_DirectCompare
 *
 * Returns T iff the two pointers are the same address.
 */
extern Boolean
List_DirectCompare(void* datum1, void* datum2);

/*
 * List_Equal
 */
extern Boolean
List_Equal(List list1, List list2, List_Comparer function);

/*
 * List_Append
 *
 * Appends list2 to the end of list1.
 */
extern void
List_Append(List list1, List list2);

/*
 * List_Push
 *
 * Adds datum to the head of the list.
 */
extern void
List_Push(List list, void* datum);

#endif
