/*  -*- mode:     C -*- --------------------------------------------------
* File:         rules/factor.c
* Description:  Implements factor.h
* Author:       Brett Kessler
* Created:      Mon Apr 13 21:34:04 1992
* Modified:     Sun May  3 15:14:04 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#include "factor.h"
#include "attribute.h"
#include "Malloc.h"
#include "print.h"
#include "item.h"
#include <stdlib.h>

Factor
Factor_CreateFromAlternative(List alternative)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_alternativeType;
  f->u.alternatives = alternative;
  return f;
}

Factor
Factor_CreateFromOption(void* /* Item */ option)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_optionalType;
  f->u.optionalItem = option;
  return f;
}

Factor
Factor_CreateFromRepetition(Factor rep)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_repetitionType;
  f->u.repetition = rep;
  return f;
}

Factor
Factor_CreateFromAVM(AVM avm)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_avmType;
  f->u.avmFeatures = avm;
  return f;
}

Factor
Factor_CreateFromIndexedAVM(IndexedAVM avm)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_indexedAVMType;
  f->u.indexedAVM = avm;
  return f;
}

Factor
Factor_CreateFromSegment(int segment)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_segmentType;
  f->u.number = segment;
  return f;
}

Factor
Factor_CreateFromEnvironbar(void)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_environbarType;
  return f;
}

Factor
Factor_CreateFromZero(void)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_zeroType;
  return f;
}

Factor
Factor_CreateFromBoundary(Factor_Boundary type)
{
  Factor f;
  f = Malloc_Struct(factor);
  switch (type) {
  case Factor_syllableBoundary: f->type = Factor_syllableBoundaryType; break;
  case Factor_utteranceBoundary: f->type = Factor_utteranceBoundaryType; break;
  default: 
    (void) fprintf(stderr, "Bad type of boundary: %d\n", type);
    exit(1);
  }
  return f;
}

static void
Factor_PrintBoundary(Factor_Boundary b, FILE* file)
{
  switch (b) {
  case Factor_syllableBoundary:
    (void) fputs("Factor_syllableBoundary", file);
    break;
  case Factor_wordBoundary:
    (void) fputs("Factor_wordBoundary", file);
    break;
  case Factor_utteranceBoundary:
    (void) fputs("Factor_utteranceBoundary", file);
    break;
  }
}

/* Inner loop for printing items in an alternative Factor. */
static void
PrintAlternatives(List items, FILE* file, void* /* Item */ focus)
{
  Cons itemCons;
  void* item;
  int count = 0;
  List_ForEach(items, itemCons, item) {
    (void) fprintf(file, ",\n");
    Print_Indent(file);
    (void) fprintf(file, "/* Factor alternative #%d */\n", count);
    Item_Print(item, file, focus);
    count++;
  }
}

/* Prints the lexical attributes associated with a word boundary. */
static void
PrintAttributes(List attributes, FILE* file)
{
  Cons aCons;
  Attribute attribute;
  Print_Indent(file);
  if (attributes == NULL) {
    (void) fprintf(file, "NULL");
    return;
  }
  Print_level++;
  (void) fprintf(file, "Attributes_Create(%d", attributes->count);
  List_ForEach(attributes, aCons, attribute) {
    (void) fprintf(file, ",\n");
    Attribute_Print(attribute, file);
  }
  (void) fprintf(file, ")");
  Print_level--;
}

void
Factor_Print(Factor f, FILE* file, void* /* Item */ focus)
{
  Print_Indent(file);
  Print_level++;
  switch (f->type) {
  case Factor_alternativeType:
    (void)
      fprintf(file,
	      "Factor_CreateFromAlternative(%d",
	      f->u.alternatives->count);
    PrintAlternatives(f->u.alternatives, file, focus);
    break;
  case Factor_optionalType:
    (void) fputs("Factor_CreateFromOption(\n", file);
    Item_Print(f->u.optionalItem, file, focus);
    break;
  case Factor_repetitionType:
    (void) fputs("Factor_CreateFromRepetition(\n", file);
    Factor_Print(f->u.repetition, file, focus);
    break;
  case Factor_avmType:
    (void) fputs("Factor_CreateFromAVM(\n", file);
    AVM_Print(f->u.avmFeatures, file);
    break;
  case Factor_indexedAVMType:
    (void) fputs("Factor_CreateFromIndexedAVM(\n", file);
    IndexedAVM_Print(f->u.indexedAVM, file);
    break;
  case Factor_segmentType:
    (void) fprintf(file, "Factor_CreateFromSegment(%d", f->u.number);
    break;
  case Factor_environbarType:
    /*
     * Instead of distinguishing the source description from the 
     * environment, the rule format within derive/ will have the
     * entire structural description in line, with focus-start and
     * focus-end delimiters surrounding the source (the part replaced
     * or modified by the structural change).
     */
    if (focus == NULL) {
      (void) fprintf(stderr, "Rule has environment bar in focus?\n");
      exit(1);
    }
    (void) fprintf(file, "Factor_CreateFromItem(Item_Create(3,\n");
    Print_level++;
    Print_Indent(file);
    (void) fprintf(file, "Factor_CreateFocusStart(),\n");
    Print_Indent(file);
    (void) fprintf(file, "Factor_CreateFromItem(");
    Item_Print(focus, file, NULL);
    (void) fputs("),\n", file);
    Print_Indent(file);
    (void) fprintf(file, "Factor_CreateFocusEnd())\n");
    Print_level--;
    break;
  case Factor_zeroType:
    (void) fputs("Factor_CreateFromZero(", file);
    break;
  case Factor_syllableBoundaryType:
    (void) fputs("Factor_CreateFromBoundary(", file);
    Factor_PrintBoundary(Factor_syllableBoundary, file);
    break;
  case Factor_wordBoundaryType:
    (void) fputs("Factor_CreateWordBoundary(", file);
    PrintAttributes(f->u.attributes, file);
    break;
  case Factor_utteranceBoundaryType:
    (void) fputs("Factor_CreateFromBoundary(", file);
    Factor_PrintBoundary(Factor_utteranceBoundary, file);
    break;
  }
  (void) fputs(")", file);
  Print_level--;
}

Factor
Factor_CreateWordBoundary(List attributes)
{
  Factor f;
  f = Malloc_Struct(factor);
  f->type = Factor_wordBoundaryType;
  f->u.attributes = attributes;
  return f;
}

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
