/*  -*- mode:     C -*- --------------------------------------------------
* File:         rules/factor.h
* Description:  Interface file.
* Author:       Brett Kessler
* Created:      Thu Apr  9 08:44:57 1992
* Modified:     Sun May  3 15:14:04 1992 (Brett Kessler)
* Language:     C
******************************************************************************
*/

#if !defined(FACTOR_H)
#define FACTOR_H

#include "avm.h"
#include "indexedAVM.h"
#include "list.h"
#include <stdio.h>

typedef enum factor_Boundary {
  Factor_syllableBoundary,
  Factor_wordBoundary,
  Factor_utteranceBoundary
  } Factor_Boundary;

typedef enum factor_Type {
  Factor_alternativeType,       /* {X, Y, Z} */
  Factor_optionalType,          /* (X) */
  Factor_repetitionType,        /* X* */
  Factor_avmType,               /* [+a, -b] */
  Factor_indexedAVMType,        /* [+a, -b]_i */
  Factor_segmentType,           /* bh */
  Factor_environbarType,        /* __ */
  Factor_zeroType,              /* 0 */
  Factor_syllableBoundaryType,  /* SB */
  Factor_wordBoundaryType,      /* WB */
  Factor_utteranceBoundaryType  /* UB */
} Factor_Type;

/*
 * Factor
 *
 * A Factor is a fairly abstract type: virtually anything other than
 * a sequence (which is an Item).  This type serves to package them into
 * a single data type, which is the sort of things found in an Item.
 */
typedef struct factor {
  Factor_Type type;
  union {
    List /* of Item */ alternatives;    /* _alternativeType */
    void* /* Item */ optionalItem;      /* _optionalType */
    AVM avmFeatures;                    /*  _avmType */
    struct factor* repetition;          /* _repetitionType */
    IndexedAVM indexedAVM;
    int number;                         /* _segmentType */
    List /* of Attribute */ attributes; /* wordBoundaryType */
  } u;
}* Factor;

/*
 * Factor_CreateFromAlternative
 *
 * Give a list of items, returns an alternatives Factor.
 */
extern Factor
Factor_CreateFromAlternative(List /* of Item */ alternative);

/*
 * Factor_CreateFromOption
 *
 * Given an item, returns an optional Factor.  The void* is here because
 * Factor is layered below Item.
 */
extern Factor
Factor_CreateFromOption(void* /* Item */ option);

/*
 * Factor_CreateFromRepetition
 *
 * Given another factor, returns a Factor that is a repetition Factor.
 */
extern Factor
Factor_CreateFromRepetition(Factor rep);

/*
 * Factor_CreateFromAVM
 *
 * Given an AVM, returns an AVM Factor.
 */
extern Factor
Factor_CreateFromAVM(AVM avm);

/*
 * Factor_CreateFromIndexedAVM
 * 
 * Given an AVM, returns an indexed AVM Factor.
 */
extern Factor
Factor_CreateFromIndexedAVM(IndexedAVM avm);

/*
 * Factor_CreateFromSegment
 *
 * Given a segment ID, returns a segment Factor.
 */
extern Factor
Factor_CreateFromSegment(int segment);

/*
 * Factor_CreateFromEnvironbar
 *
 * Creates a Factor representing an environment bar (__)
 */
extern Factor
Factor_CreateFromEnvironbar(void);

/*
 * Factor_CreateFromZero
 *
 * Creates a Factor representing a zero (null Factor, "0")
 */
extern Factor
Factor_CreateFromZero(void);

/*
 * Factor_CreateWordBoundary
 *
 * Given a list of lexical attributes (cited by ID number), returns a
 * word boundary Factor.  The attributes List may be NULL.
 */
extern Factor
Factor_CreateWordBoundary(List attributes);

/*
 * Factor_CreateFromBoundary
 *
 * Creates a Factor for representing the type of boundary named in the
 * argument.
 */
extern Factor
Factor_CreateFromBoundary(Factor_Boundary type);

/*
 * Factor_Print
 *
 * Prints on the file (rules.c) the C code necessary to rebuild the
 * Factor structure (in "derive").  The focus argument gives the
 * source definition in the parent rule, which is substituted (with
 * appropriate delimiters) for any environment bar that might be 
 * encountered within this factor.
 */
void
Factor_Print(Factor f, FILE* file, void* /* Item */ focus);

#endif  /* end of code */

/*
******************************************************************************
*                                End of file
******************************************************************************
*/
