%{

/* rules/readRulesYacc.y.
 *
 * Implements readRulesYacc.h: reads a RULES file on standard input,
 * parses it, and returns a ruleSystem data structure.
 */

#include "attribute.h"
#include "avm.h"
#include "factor.h"
#include "feature.h"
#include "indexedAVM.h"
#include "item.h"
#include "list.h"
#include "plainSignedFeature.h"
#include "rule.h"
#include "rules.h"
#include "ruleSystem.h"
#include "sign.h"
#include "String.h"
#include "readRulesYacc.h"
#include <stdlib.h>

static RuleSystem 
ruleSystem = NULL;

%}

%union {
  Attribute attribute;
  AVM avm;
  int integer;
  char character;
  Factor factor;
  Feature feature;
  IndexedAVM indexedAVM;
  Item item;
  List list;
  PlainSignedFeature plainSignedFeature;
  Rule rule;
  RuleSystem ruleSystem;
  Rules rules;
  Sign sign;
  Sign_Variable sign_variable;
  char* string;
}

%token <integer> ALPHA
%token <integer> ATTRIBUTE_NAME
%token <integer> BECOMES
%token <integer> BETA
%token <integer> COMMA
%token <integer> DELTA
%token <integer> END_ALTERNATIVE
%token <integer> END_AVM
%token <integer> END_OPTIONAL
%token <integer> END_WORD_FEATURES
%token <integer> ENVIRONBAR
%token <integer> EQUALS
%token <integer> FEATURE_NAME
%token <integer> GAMMA
%token <integer> IN
%token <integer> INDEX
%token <integer> MAXIMIZE_ONSETS
%token <integer> MINUS
%token <integer> NAME
%token <integer> ONSETS
%token <integer> PLUS
%token <integer> RULE_TERMINATOR
%token <integer> RULES
%token <integer> SB
%token <integer> SEGMENT_NAME
%token <integer> STAR
%token <integer> START_ALTERNATIVE
%token <integer> START_AVM
%token <integer> START_OPTIONAL
%token <integer> START_WORD_FEATURES
%token <integer> UB
%token <integer> WB
%token <integer> WORD1
%token <integer> WORD2
%token <integer> ZERO

%type <list> alternative
%type <list> alternativeSignedFeature
%type <list> alternatives
%type <attribute> attr
%type <list> attrs
%type <list> attributes
%type <avm> avm
%type <sign> constantSign
%type <item> environ
%type <factor> factor
%type <integer> featName
%type <list> features
%type <sign_variable> greek
%type <character> index
%type <indexedAVM> indexed_avm
%type <item> item
%type <string> name
%type <item> onsets
%type <item> optional
%type <plainSignedFeature> plainSignedFeature
%type <factor> rep
%type <rule> rule
%type <rules> rules
%type <feature> signedFeature
%type <integer> segment
%type <sign> sign
%type <list> signedFeatureAlternatives
%type <integer> specialRule
%type <factor> wordBoundary

%%

file:		  onsets RULES rules
                  { ruleSystem = RuleSystem_Create($1, $3); }
		;

onsets:           ONSETS item
                  { $$ = $2; }
                ;

rules:		  rule
                  { $$ = Rules_Create($1); }
		| rules rule
                  { $$ = Rules_Add($1, $2); }
		;

rule:		  name EQUALS item BECOMES item environ RULE_TERMINATOR
                  { $$ = Rule_CreateOrdinary($1, $3, $5, $6); }
		| specialRule RULE_TERMINATOR
                  { $$ = Rule_CreateSpecial($1); }
		;

name:             NAME
                  { $$ = Strdup((char*) yytext); }
		;

specialRule:	  MAXIMIZE_ONSETS
                  { $$ = Rule_maximizeOnsets; }
		;

environ:	  /* empty */
                  { $$ = NULL; }
		| IN item
                  { $$ = $2; }
		;

item:		  factor
                  { $$ = Item_Create($1); }
		| item factor
                  { $$ = Item_Add($1, $2); }
		;

factor:		  alternative
                  { $$ = Factor_CreateFromAlternative($1); }
		| optional
                  { $$ = Factor_CreateFromOption($1); }
		| rep
                  { $$ = Factor_CreateFromRepetition($1); }
		| avm
                  { $$ = Factor_CreateFromAVM($1); }
		| indexed_avm
                  { $$ = Factor_CreateFromIndexedAVM($1); }
		| segment
                  { $$ = Factor_CreateFromSegment($1); }
		| ENVIRONBAR
                  { $$ = Factor_CreateFromEnvironbar(); }
		| ZERO
                  { $$ = Factor_CreateFromZero(); }
		| wordBoundary
                  { $$ = $1; }
		| SB
                  { $$ = Factor_CreateFromBoundary(Factor_syllableBoundary); }
		| UB
                  { $$ = Factor_CreateFromBoundary(Factor_utteranceBoundary); }
		;

wordBoundary:     WB attributes
                  { $$ = Factor_CreateWordBoundary($2); }
                | WB
                  { $$ = Factor_CreateWordBoundary(NULL); }
                ;

attributes:       START_WORD_FEATURES attrs END_WORD_FEATURES
                  { $$ = $2; }
                ;

attrs:            attr
                  { $$ = List_Add(NULL, $1); }
                | attrs attr
                  { $$ = List_Add($1, $2); }
                ;

attr:             constantSign ATTRIBUTE_NAME
                  { $$ = Attribute_Create($1, $2); }
                ;

constantSign:     PLUS
                  { $$ = Sign_CreateConstant(Sign_PLUS); }
		| MINUS
                  { $$ = Sign_CreateConstant(Sign_MINUS); }
                ;

segment:          SEGMENT_NAME
                  { $$ = $1; }
                ;

alternative:	  START_ALTERNATIVE alternatives END_ALTERNATIVE
                  { $$ = $2; }
		;

alternatives:	  item
                  { $$ = List_Add(NULL, $1); }
		| alternatives COMMA item
                  { $$ = List_Add($1, $3); }
		;

optional:	  START_OPTIONAL item END_OPTIONAL
                  { $$ = $2; }
		;

avm:		  START_AVM features END_AVM
                  { $$ = AVM_Create($2); }
		| START_AVM END_AVM
                  { $$ = AVM_Create(NULL); }
		;

features:	  signedFeature
                  { $$ = List_Add(NULL, $1); }
		| features signedFeature
                  { $$ = List_Add($1, $2); }
		| features COMMA signedFeature
                  { $$ = List_Add($1, $3); }
		;

signedFeature:	  plainSignedFeature
                  { $$ = Feature_CreateFromPlain($1); }
		| alternativeSignedFeature
                  { $$ = Feature_CreateFromAlternatives($1); }
		;

plainSignedFeature: sign featName
                  { $$ = PlainSignedFeature_Create($1, $2); }
		;

featName:          FEATURE_NAME
                   { $$ = $1; }
                ;

alternativeSignedFeature: START_ALTERNATIVE 
		  signedFeatureAlternatives 
		  END_ALTERNATIVE
                  { $$ = $2; }
		;

signedFeatureAlternatives: plainSignedFeature
                  { $$ = List_Add(NULL, $1); }
		| signedFeatureAlternatives COMMA plainSignedFeature
                  { $$ = List_Add($1, $3); }
		;

sign:		  constantSign
                  { $$ = $1; }
		| greek
                  { $$ = Sign_CreateVariable($1); }
		;

greek:		  ALPHA
                  { $$ = Sign_ALPHA; }
		| BETA
                  { $$ = Sign_BETA; }
		| GAMMA
                  { $$ = Sign_GAMMA; }
		| DELTA
                  { $$ = Sign_DELTA; }
		;

indexed_avm:	  avm index
                  { $$ = IndexedAVM_Create($1, $2); }
		;

index:		  INDEX
                  { $$ = yytext[1]; }
		;

rep:		  factor STAR
                  { $$ = $1; }
		;

%%

#include "readRulesLex.c"

int
ReadRulesYacc_Position(void)
{
  return (int) ftell(yyin);
}

int
yyerror(char* s)
{
  int position;
  position = ReadRulesYacc_Position();
  fprintf(stderr, "%s in file", s);
  if (position >= 0) fprintf(stderr, ", near position %d\n", position);
  else putchar('\n');
  exit(1);
}

RuleSystem
ReadRulesYacc_Parse()
{
  yyparse();
  return ruleSystem;
}
