%{

/*
 * segments/readSegmentsYacc.y
 *
 * Parses stdin (the SEGMENTS file) and writes 6 code fragment files.
 * The information is written out as soon as it is recognized, without
 * ever building up a data structure for the entire file.
 */

#include "Malloc.h"
#include "readSegmentsYacc.h"
#include <string.h>
#include <stdio.h>

static FILE* segmentsLex = NULL;
static FILE* segmentsH = NULL;
static FILE* segmentsC = NULL;
static FILE* attributesLex = NULL;
static FILE* attributesH = NULL;
static FILE* attributesC = NULL;
static int nSegments = 0;
static int nAttributes = 0;

typedef int Features_Id;
#define Features_NONE -1
#include "../frag/featuresH"
#include "../frag/featuresC"
%}

%union {
  int integer;
  unsigned long mask;
  char* token;
}

%token <integer> ATTRIBUTES
%token <integer> FEATURE_NAME
%token <integer> ROW_END
%token <integer> EQUALS
%token <token>   TOKEN
%token <integer> SEGMENTS

%type <token> ascii
%type <mask>  feature
%type <mask>  features
%type <token> latex

%%

file:			  segments attributesPart
			;

segments:		  SEGMENTS
                          { (void)
			      fputs("struct Segment_fact"
				    " Segments_Facts[Segments_COUNT + 1]"
				    " = {\n", 
				    segmentsC);
			  }
                          definitions
                          { (void)
			      fprintf(segmentsH,
				      "#define Segments_COUNT %d\n",
				      nSegments);
			    (void) 
			      fputs("  {\"ILLEGAL\", \"ILLEGAL\", 0}\n};"
				    "\n\n", segmentsC);
			  }
			;

definitions:		  definition
			| definitions definition
			;

definition:		  ascii latex EQUALS features ROW_END
			  { 
			    (void)
			      fprintf(segmentsLex, 
				      "\"%s\"\t\t{ yylval.integer = %d;"
				      " return SEGMENT_NAME; }\n",
				      $1, nSegments);
			    (void) 
			      fprintf(segmentsC,
				      "  {\"%s\", \"%s\", %d}"
				      ", /* %d */\n", 
				      $1, 
				      $2,
				      $4,
				      nSegments);
			    nSegments++;
			  }
			;

ascii:			  TOKEN
			  { $$ = $1; }
			;

latex:			  TOKEN
			  { $$ = $1; }
			;

features:		  feature
                          { $$ = $1; }
			| features feature
                          { $$ = $1 + $2; }
			;

feature:		  FEATURE_NAME
			  { $$ = Features_mask[$1]; }
			;

attributesPart:		  ATTRIBUTES
                          { (void)
			      fputs("char* Attributes_Names[Attributes_COUNT"
				    " + 1] = {\n", attributesC);
			  }
                          attributes
                          { (void)
			      fprintf(attributesH,
				      "#define Attributes_COUNT %d\n",
				      nAttributes);
			    (void) fputs("  \"ILLEGAL\"\n};\n\n", attributesC);
			  }
			;

attributes:		  attribute
			| attributes attribute
			;

attribute:		  TOKEN
                          { (void)
			      fprintf(attributesLex, 
				      "\"%s\"\t{ yylval.integer = %d;"
				      " return ATTRIBUTE_NAME; }\n",
				      $1, nAttributes); 
			    (void)
			      fprintf(attributesC,
				      "  \"%s\",\t/* %d */\n",
				      $1, nAttributes);
			    nAttributes++;
			  }

			;

%%

#include "readSegmentsLex.c"

int
ReadSegmentsYacc_Position(void)
{
  return (int) ftell(yyin);
}

int yywrap(void)
{
        return 1;
} 

int
yyerror(char* s)
{
  int position;
  position = ReadSegmentsYacc_Position();
  fprintf(stderr, "%s in file", s);
  if (position >= 0) fprintf(stderr, ", near position %d\n", position);
  else putchar('\n');
  exit(1);
}

void
ReadSegmentsYacc_Parse(FILE* segmentsLexFile, 
		       FILE* segmentsHFile,
		       FILE* segmentsCFile,
		       FILE* attributesLexFile,
		       FILE* attributesHFile,
		       FILE* attributesCFile)
{
  segmentsLex = segmentsLexFile;
  segmentsH = segmentsHFile;
  segmentsC = segmentsCFile;
  attributesLex = attributesLexFile;
  attributesH = attributesHFile;
  attributesC = attributesCFile;
  yyparse();
}

